forked from openwrt/openwrt
-
Notifications
You must be signed in to change notification settings - Fork 0
Commit
This commit does not belong to any branch on this repository, and may belong to a fork outside of the repository.
mac80211: backport upstream fixes for FragAttacks
From the patch series description: Several security issues in the 802.11 implementations were found by Mathy Vanhoef (New York University Abu Dhabi), who has published all the details at https://papers.mathyvanhoef.com/usenix2021.pdf Specifically, the following CVEs were assigned: * CVE-2020-24586 - Fragmentation cache not cleared on reconnection * CVE-2020-24587 - Reassembling fragments encrypted under different keys * CVE-2020-24588 - Accepting non-SPP A-MSDU frames, which leads to payload being parsed as an L2 frame under an A-MSDU bit toggling attack * CVE-2020-26139 - Forwarding EAPOL from unauthenticated sender * CVE-2020-26140 - Accepting plaintext data frames in protected networks * CVE-2020-26141 - Not verifying TKIP MIC of fragmented frames * CVE-2020-26142 - Processing fragmented frames as full frames * CVE-2020-26143 - Accepting fragmented plaintext frames in protected networks * CVE-2020-26144 - Always accepting unencrypted A-MSDU frames that start with RFC1042 header with EAPOL ethertype * CVE-2020-26145 - Accepting plaintext broadcast fragments as full frames * CVE-2020-26146 - Reassembling encrypted fragments with non-consecutive packet numbers * CVE-2020-26147 - Reassembling mixed encrypted/plaintext fragments In general, the scope of these attacks is that they may allow an attacker to * inject L2 frames that they can more or less control (depending on the vulnerability and attack method) into an otherwise protected network; * exfiltrate (some) network data under certain conditions, this is specific to the fragmentation issues. A subset of these issues is known to apply to the Linux IEEE 802.11 implementation (mac80211). Where it is affected, the attached patches fix the issues, even if not all of them reference the exact CVE IDs. In addition, driver and/or firmware updates may be necessary, as well as potentially more fixes to mac80211, depending on how drivers are using it. Specifically, for Intel devices, firmware needs to be updated to the most recently released versions (which was done without any reference to the security issues) to address some of the vulnerabilities. To have a single set of patches, I'm also including patches for the ath10k and ath11k drivers here. We currently don't have information about how other drivers are, if at all, affected. Signed-off-by: Felix Fietkau <nbd@nbd.name>
- Loading branch information
Showing
23 changed files
with
1,466 additions
and
7 deletions.
There are no files selected for viewing
This file contains bidirectional Unicode text that may be interpreted or compiled differently than what appears below. To review, open the file in an editor that reveals hidden Unicode characters.
Learn more about bidirectional Unicode characters
180 changes: 180 additions & 0 deletions
180
...ernel/mac80211/patches/ath/300-ath10k-add-CCMP-PN-replay-protection-for-fragmented-.patch
This file contains bidirectional Unicode text that may be interpreted or compiled differently than what appears below. To review, open the file in an editor that reveals hidden Unicode characters.
Learn more about bidirectional Unicode characters
Original file line number | Diff line number | Diff line change |
---|---|---|
@@ -0,0 +1,180 @@ | ||
From: Wen Gong <wgong@codeaurora.org> | ||
Date: Tue, 11 May 2021 20:02:52 +0200 | ||
Subject: [PATCH] ath10k: add CCMP PN replay protection for fragmented | ||
frames for PCIe | ||
|
||
PN replay check for not fragmented frames is finished in the firmware, | ||
but this was not done for fragmented frames when ath10k is used with | ||
QCA6174/QCA6377 PCIe. mac80211 has the function | ||
ieee80211_rx_h_defragment() for PN replay check for fragmented frames, | ||
but this does not get checked with QCA6174 due to the | ||
ieee80211_has_protected() condition not matching the cleared Protected | ||
bit case. | ||
|
||
Validate the PN of received fragmented frames within ath10k when CCMP is | ||
used and drop the fragment if the PN is not correct (incremented by | ||
exactly one from the previous fragment). This applies only for | ||
QCA6174/QCA6377 PCIe. | ||
|
||
Tested-on: QCA6174 hw3.2 PCI WLAN.RM.4.4.1-00110-QCARMSWP-1 | ||
|
||
Cc: stable@vger.kernel.org | ||
Signed-off-by: Wen Gong <wgong@codeaurora.org> | ||
Signed-off-by: Jouni Malinen <jouni@codeaurora.org> | ||
Signed-off-by: Johannes Berg <johannes.berg@intel.com> | ||
--- | ||
|
||
--- a/drivers/net/wireless/ath/ath10k/htt.h | ||
+++ b/drivers/net/wireless/ath/ath10k/htt.h | ||
@@ -846,6 +846,7 @@ enum htt_security_types { | ||
|
||
#define ATH10K_HTT_TXRX_PEER_SECURITY_MAX 2 | ||
#define ATH10K_TXRX_NUM_EXT_TIDS 19 | ||
+#define ATH10K_TXRX_NON_QOS_TID 16 | ||
|
||
enum htt_security_flags { | ||
#define HTT_SECURITY_TYPE_MASK 0x7F | ||
--- a/drivers/net/wireless/ath/ath10k/htt_rx.c | ||
+++ b/drivers/net/wireless/ath/ath10k/htt_rx.c | ||
@@ -1746,16 +1746,87 @@ static void ath10k_htt_rx_h_csum_offload | ||
msdu->ip_summed = ath10k_htt_rx_get_csum_state(msdu); | ||
} | ||
|
||
+static u64 ath10k_htt_rx_h_get_pn(struct ath10k *ar, struct sk_buff *skb, | ||
+ u16 offset, | ||
+ enum htt_rx_mpdu_encrypt_type enctype) | ||
+{ | ||
+ struct ieee80211_hdr *hdr; | ||
+ u64 pn = 0; | ||
+ u8 *ehdr; | ||
+ | ||
+ hdr = (struct ieee80211_hdr *)(skb->data + offset); | ||
+ ehdr = skb->data + offset + ieee80211_hdrlen(hdr->frame_control); | ||
+ | ||
+ if (enctype == HTT_RX_MPDU_ENCRYPT_AES_CCM_WPA2) { | ||
+ pn = ehdr[0]; | ||
+ pn |= (u64)ehdr[1] << 8; | ||
+ pn |= (u64)ehdr[4] << 16; | ||
+ pn |= (u64)ehdr[5] << 24; | ||
+ pn |= (u64)ehdr[6] << 32; | ||
+ pn |= (u64)ehdr[7] << 40; | ||
+ } | ||
+ return pn; | ||
+} | ||
+ | ||
+static bool ath10k_htt_rx_h_frag_pn_check(struct ath10k *ar, | ||
+ struct sk_buff *skb, | ||
+ u16 peer_id, | ||
+ u16 offset, | ||
+ enum htt_rx_mpdu_encrypt_type enctype) | ||
+{ | ||
+ struct ath10k_peer *peer; | ||
+ union htt_rx_pn_t *last_pn, new_pn = {0}; | ||
+ struct ieee80211_hdr *hdr; | ||
+ bool more_frags; | ||
+ u8 tid, frag_number; | ||
+ u32 seq; | ||
+ | ||
+ peer = ath10k_peer_find_by_id(ar, peer_id); | ||
+ if (!peer) { | ||
+ ath10k_dbg(ar, ATH10K_DBG_HTT, "invalid peer for frag pn check\n"); | ||
+ return false; | ||
+ } | ||
+ | ||
+ hdr = (struct ieee80211_hdr *)(skb->data + offset); | ||
+ if (ieee80211_is_data_qos(hdr->frame_control)) | ||
+ tid = ieee80211_get_tid(hdr); | ||
+ else | ||
+ tid = ATH10K_TXRX_NON_QOS_TID; | ||
+ | ||
+ last_pn = &peer->frag_tids_last_pn[tid]; | ||
+ new_pn.pn48 = ath10k_htt_rx_h_get_pn(ar, skb, offset, enctype); | ||
+ more_frags = ieee80211_has_morefrags(hdr->frame_control); | ||
+ frag_number = le16_to_cpu(hdr->seq_ctrl) & IEEE80211_SCTL_FRAG; | ||
+ seq = (__le16_to_cpu(hdr->seq_ctrl) & IEEE80211_SCTL_SEQ) >> 4; | ||
+ | ||
+ if (frag_number == 0) { | ||
+ last_pn->pn48 = new_pn.pn48; | ||
+ peer->frag_tids_seq[tid] = seq; | ||
+ } else { | ||
+ if (seq != peer->frag_tids_seq[tid]) | ||
+ return false; | ||
+ | ||
+ if (new_pn.pn48 != last_pn->pn48 + 1) | ||
+ return false; | ||
+ | ||
+ last_pn->pn48 = new_pn.pn48; | ||
+ } | ||
+ | ||
+ return true; | ||
+} | ||
+ | ||
static void ath10k_htt_rx_h_mpdu(struct ath10k *ar, | ||
struct sk_buff_head *amsdu, | ||
struct ieee80211_rx_status *status, | ||
bool fill_crypt_header, | ||
u8 *rx_hdr, | ||
- enum ath10k_pkt_rx_err *err) | ||
+ enum ath10k_pkt_rx_err *err, | ||
+ u16 peer_id, | ||
+ bool frag) | ||
{ | ||
struct sk_buff *first; | ||
struct sk_buff *last; | ||
- struct sk_buff *msdu; | ||
+ struct sk_buff *msdu, *temp; | ||
struct htt_rx_desc *rxd; | ||
struct ieee80211_hdr *hdr; | ||
enum htt_rx_mpdu_encrypt_type enctype; | ||
@@ -1768,6 +1839,7 @@ static void ath10k_htt_rx_h_mpdu(struct | ||
bool is_decrypted; | ||
bool is_mgmt; | ||
u32 attention; | ||
+ bool frag_pn_check = true; | ||
|
||
if (skb_queue_empty(amsdu)) | ||
return; | ||
@@ -1866,6 +1938,24 @@ static void ath10k_htt_rx_h_mpdu(struct | ||
} | ||
|
||
skb_queue_walk(amsdu, msdu) { | ||
+ if (frag && !fill_crypt_header && is_decrypted && | ||
+ enctype == HTT_RX_MPDU_ENCRYPT_AES_CCM_WPA2) | ||
+ frag_pn_check = ath10k_htt_rx_h_frag_pn_check(ar, | ||
+ msdu, | ||
+ peer_id, | ||
+ 0, | ||
+ enctype); | ||
+ | ||
+ if (!frag_pn_check) { | ||
+ /* Discard the fragment with invalid PN */ | ||
+ temp = msdu->prev; | ||
+ __skb_unlink(msdu, amsdu); | ||
+ dev_kfree_skb_any(msdu); | ||
+ msdu = temp; | ||
+ frag_pn_check = true; | ||
+ continue; | ||
+ } | ||
+ | ||
ath10k_htt_rx_h_csum_offload(msdu); | ||
ath10k_htt_rx_h_undecap(ar, msdu, status, first_hdr, enctype, | ||
is_decrypted); | ||
@@ -2071,7 +2161,8 @@ static int ath10k_htt_rx_handle_amsdu(st | ||
ath10k_htt_rx_h_unchain(ar, &amsdu, &drop_cnt, &unchain_cnt); | ||
|
||
ath10k_htt_rx_h_filter(ar, &amsdu, rx_status, &drop_cnt_filter); | ||
- ath10k_htt_rx_h_mpdu(ar, &amsdu, rx_status, true, first_hdr, &err); | ||
+ ath10k_htt_rx_h_mpdu(ar, &amsdu, rx_status, true, first_hdr, &err, 0, | ||
+ false); | ||
msdus_to_queue = skb_queue_len(&amsdu); | ||
ath10k_htt_rx_h_enqueue(ar, &amsdu, rx_status); | ||
|
||
@@ -3027,7 +3118,7 @@ static int ath10k_htt_rx_in_ord_ind(stru | ||
ath10k_htt_rx_h_ppdu(ar, &amsdu, status, vdev_id); | ||
ath10k_htt_rx_h_filter(ar, &amsdu, status, NULL); | ||
ath10k_htt_rx_h_mpdu(ar, &amsdu, status, false, NULL, | ||
- NULL); | ||
+ NULL, peer_id, frag); | ||
ath10k_htt_rx_h_enqueue(ar, &amsdu, status); | ||
break; | ||
case -EAGAIN: |
66 changes: 66 additions & 0 deletions
66
...ge/kernel/mac80211/patches/ath/301-ath10k-drop-fragments-with-multicast-DA-for-PCIe.patch
This file contains bidirectional Unicode text that may be interpreted or compiled differently than what appears below. To review, open the file in an editor that reveals hidden Unicode characters.
Learn more about bidirectional Unicode characters
Original file line number | Diff line number | Diff line change |
---|---|---|
@@ -0,0 +1,66 @@ | ||
From: Wen Gong <wgong@codeaurora.org> | ||
Date: Tue, 11 May 2021 20:02:53 +0200 | ||
Subject: [PATCH] ath10k: drop fragments with multicast DA for PCIe | ||
|
||
Fragmentation is not used with multicast frames. Discard unexpected | ||
fragments with multicast DA. This fixes CVE-2020-26145. | ||
|
||
Tested-on: QCA6174 hw3.2 PCI WLAN.RM.4.4.1-00110-QCARMSWP-1 | ||
|
||
Cc: stable@vger.kernel.org | ||
Signed-off-by: Wen Gong <wgong@codeaurora.org> | ||
Signed-off-by: Jouni Malinen <jouni@codeaurora.org> | ||
Signed-off-by: Johannes Berg <johannes.berg@intel.com> | ||
--- | ||
|
||
--- a/drivers/net/wireless/ath/ath10k/htt_rx.c | ||
+++ b/drivers/net/wireless/ath/ath10k/htt_rx.c | ||
@@ -2617,6 +2617,13 @@ static bool ath10k_htt_rx_proc_rx_frag_i | ||
rx_desc = (struct htt_hl_rx_desc *)(skb->data + tot_hdr_len); | ||
rx_desc_info = __le32_to_cpu(rx_desc->info); | ||
|
||
+ hdr = (struct ieee80211_hdr *)((u8 *)rx_desc + rx_hl->fw_desc.len); | ||
+ | ||
+ if (is_multicast_ether_addr(hdr->addr1)) { | ||
+ /* Discard the fragment with multicast DA */ | ||
+ goto err; | ||
+ } | ||
+ | ||
if (!MS(rx_desc_info, HTT_RX_DESC_HL_INFO_ENCRYPTED)) { | ||
spin_unlock_bh(&ar->data_lock); | ||
return ath10k_htt_rx_proc_rx_ind_hl(htt, &resp->rx_ind_hl, skb, | ||
@@ -2624,8 +2631,6 @@ static bool ath10k_htt_rx_proc_rx_frag_i | ||
HTT_RX_NON_TKIP_MIC); | ||
} | ||
|
||
- hdr = (struct ieee80211_hdr *)((u8 *)rx_desc + rx_hl->fw_desc.len); | ||
- | ||
if (ieee80211_has_retry(hdr->frame_control)) | ||
goto err; | ||
|
54 changes: 54 additions & 0 deletions
54
...ernel/mac80211/patches/ath/303-ath10k-drop-MPDU-which-has-discard-flag-set-by-firmw.patch
This file contains bidirectional Unicode text that may be interpreted or compiled differently than what appears below. To review, open the file in an editor that reveals hidden Unicode characters.
Learn more about bidirectional Unicode characters
Original file line number | Diff line number | Diff line change |
---|---|---|
@@ -0,0 +1,54 @@ | ||
From: Wen Gong <wgong@codeaurora.org> | ||
Date: Tue, 11 May 2021 20:02:55 +0200 | ||
Subject: [PATCH] ath10k: drop MPDU which has discard flag set by firmware | ||
for SDIO | ||
|
||
When the discard flag is set by the firmware for an MPDU, it should be | ||
dropped. This allows a mitigation for CVE-2020-24588 to be implemented | ||
in the firmware. | ||
|
||
Tested-on: QCA6174 hw3.2 SDIO WLAN.RMH.4.4.1-00049 | ||
|
||
Cc: stable@vger.kernel.org | ||
Signed-off-by: Wen Gong <wgong@codeaurora.org> | ||
Signed-off-by: Jouni Malinen <jouni@codeaurora.org> | ||
Signed-off-by: Johannes Berg <johannes.berg@intel.com> | ||
--- | ||
|
||
--- a/drivers/net/wireless/ath/ath10k/htt_rx.c | ||
+++ b/drivers/net/wireless/ath/ath10k/htt_rx.c | ||
@@ -2312,6 +2312,11 @@ static bool ath10k_htt_rx_proc_rx_ind_hl | ||
fw_desc = &rx->fw_desc; | ||
rx_desc_len = fw_desc->len; | ||
|
||
+ if (fw_desc->u.bits.discard) { | ||
+ ath10k_dbg(ar, ATH10K_DBG_HTT, "htt discard mpdu\n"); | ||
+ goto err; | ||
+ } | ||
+ | ||
/* I have not yet seen any case where num_mpdu_ranges > 1. | ||
* qcacld does not seem handle that case either, so we introduce the | ||
* same limitiation here as well. | ||
--- a/drivers/net/wireless/ath/ath10k/rx_desc.h | ||
+++ b/drivers/net/wireless/ath/ath10k/rx_desc.h | ||
@@ -1282,7 +1282,19 @@ struct fw_rx_desc_base { | ||
#define FW_RX_DESC_UDP (1 << 6) | ||
|
||
struct fw_rx_desc_hl { | ||
- u8 info0; | ||
+ union { | ||
+ struct { | ||
+ u8 discard:1, | ||
+ forward:1, | ||
+ any_err:1, | ||
+ dup_err:1, | ||
+ reserved:1, | ||
+ inspect:1, | ||
+ extension:2; | ||
+ } bits; | ||
+ u8 info0; | ||
+ } u; | ||
+ | ||
u8 version; | ||
u8 len; | ||
u8 flags; |
Oops, something went wrong.
--- a/drivers/net/wireless/ath/ath10k/htt_rx.c | ||
+++ b/drivers/net/wireless/ath/ath10k/htt_rx.c | ||
@@ -1768,6 +1768,16 @@ static u64 ath10k_htt_rx_h_get_pn(struct | ||
return pn; | ||
} | ||
|
||
+static bool ath10k_htt_rx_h_frag_multicast_check(struct ath10k *ar, | ||
+ struct sk_buff *skb, | ||
+ u16 offset) | ||
+{ | ||
+ struct ieee80211_hdr *hdr; | ||
+ | ||
+ hdr = (struct ieee80211_hdr *)(skb->data + offset); | ||
+ return !is_multicast_ether_addr(hdr->addr1); | ||
+} | ||
+ | ||
static bool ath10k_htt_rx_h_frag_pn_check(struct ath10k *ar, | ||
struct sk_buff *skb, | ||
u16 peer_id, | ||
@@ -1839,7 +1849,7 @@ static void ath10k_htt_rx_h_mpdu(struct | ||
bool is_decrypted; | ||
bool is_mgmt; | ||
u32 attention; | ||
- bool frag_pn_check = true; | ||
+ bool frag_pn_check = true, multicast_check = true; | ||
|
||
if (skb_queue_empty(amsdu)) | ||
return; | ||
@@ -1946,13 +1956,20 @@ static void ath10k_htt_rx_h_mpdu(struct | ||
0, | ||
enctype); | ||
|
||
- if (!frag_pn_check) { | ||
- /* Discard the fragment with invalid PN */ | ||
+ if (frag) | ||
+ multicast_check = ath10k_htt_rx_h_frag_multicast_check(ar, | ||
+ msdu, | ||
+ 0); | ||
+ | ||
+ if (!frag_pn_check || !multicast_check) { | ||
+ /* Discard the fragment with invalid PN or multicast DA | ||
+ */ | ||
temp = msdu->prev; | ||
__skb_unlink(msdu, amsdu); | ||
dev_kfree_skb_any(msdu); | ||
msdu = temp; | ||
frag_pn_check = true; | ||
+ multicast_check = true; | ||
continue; | ||
} | ||
|
40 changes: 40 additions & 0 deletions
40
...ge/kernel/mac80211/patches/ath/302-ath10k-drop-fragments-with-multicast-DA-for-SDIO.patch
This file contains bidirectional Unicode text that may be interpreted or compiled differently than what appears below. To review, open the file in an editor that reveals hidden Unicode characters.
Learn more about bidirectional Unicode characters
Original file line number | Diff line number | Diff line change |
---|---|---|
@@ -0,0 +1,40 @@ | ||
From: Wen Gong <wgong@codeaurora.org> | ||
Date: Tue, 11 May 2021 20:02:54 +0200 | ||
Subject: [PATCH] ath10k: drop fragments with multicast DA for SDIO | ||
|
||
Fragmentation is not used with multicast frames. Discard unexpected | ||
fragments with multicast DA. This fixes CVE-2020-26145. | ||
|
||
Tested-on: QCA6174 hw3.2 SDIO WLAN.RMH.4.4.1-00049 | ||
|
||
Cc: stable@vger.kernel.org | ||
Signed-off-by: Wen Gong <wgong@codeaurora.org> | ||
Signed-off-by: Jouni Malinen <jouni@codeaurora.org> | ||
Signed-off-by: Johannes Berg <johannes.berg@intel.com> | ||
--- | ||
|
||